import path from "node:path";
import { type ParsedComponent, parsePlaybillComponent } from "define";
import slugify from "slugify";
import YAML, { YAMLParseError } from "yaml";
import { ZodError } from "zod";
import { loadFileOrFail } from "#util";
import { MalformedResourceFileError } from "./errors";
import { extractFrontmatter, extractMarkdown } from "./markdown";

type FileFormat = "yaml" | "markdown";

function parseYAMLResourceFile(
  filePath: string,
  text: string,
): ParsedComponent[] {
  const parsedDocs = YAML.parseAllDocuments(text);

  if (parsedDocs.some((doc) => doc.toJS() === null)) {
    throw new MalformedResourceFileError("Encountered NULL resource", filePath);
  }

  const errors = parsedDocs.flatMap((doc) => doc.errors);

  if (errors.length > 0) {
    throw new MalformedResourceFileError(
      "Error parsing YAML resource",
      filePath,
      errors.map((e) => e.message).join(", "),
    );
  }

  const collection: ParsedComponent[] = [];

  for (const doc of parsedDocs) {
    const raw = doc.toJS();
    collection.push(parsePlaybillComponent(raw));
  }

  return collection;
}

function parseMarkdownResourceFile(
  filePath: string,
  text: string,
): ParsedComponent[] {
  try {
    const defaultName = path.basename(filePath, ".md");
    const defaultId = slugify(defaultName, { lower: true });
    const frontmatter = extractFrontmatter(text);
    const markdown = extractMarkdown(text);

    const together = parsePlaybillComponent({
      // Use the file name, allow it to be overridden by frontmatter
      name: defaultName,
      id: defaultId,
      ...frontmatter,
      description: markdown,
    });

    return [together];
  } catch (e) {
    if (e instanceof YAMLParseError) {
      throw new MalformedResourceFileError(
        "Error parsing Markdown frontmatter",
        filePath,
        e.message,
      );
    }

    if (e instanceof ZodError) {
      throw new MalformedResourceFileError(
        "Error parsing resource file",
        filePath,
        e.message,
      );
    }

    throw e;
  }
}

export class ComponentFile {
  private _filePath: string;
  private _raw = "";
  private _format: FileFormat;
  private _components: ParsedComponent[] = [];

  constructor(filePath: string) {
    this._filePath = path.resolve(filePath);

    const extension = path.extname(filePath).slice(1).toLowerCase();

    switch (extension) {
      case "yaml":
        this._format = "yaml";
        break;
      case "md":
        this._format = "markdown";
        break;
      default:
        throw new Error(`Unsupported file format: ${extension}`);
    }
  }

  async load() {
    this._raw = await loadFileOrFail(this._filePath);
    switch (this._format) {
      case "yaml":
        this._components = parseYAMLResourceFile(this._filePath, this._raw);
        break;
      case "markdown":
        this._components = parseMarkdownResourceFile(this._filePath, this._raw);
    }
  }

  /**
   * An array of well-formed components defined in this file
   */
  get components() {
    return this._components;
  }

  /**
   * The aboslute file path of this file
   */
  get path() {
    return this._filePath;
  }
}
