import type { MarkedExtension } from "marked";
import { marked } from "marked";
import YAML from "yaml";

const plugin: MarkedExtension = {
  renderer: {
    heading({ tokens, depth }) {
      const text = this.parser.parseInline(tokens);
      const level = Math.max(depth, 3);
      return `<h${level}>${text}</h${level}>`;
    },
  },
};

marked.use(plugin);

/**
 * Given a string of markdown, convert it to HTML
 *
 * @param markdown The markdown to convert
 * @returns A promise resolving to the HTML representation of the markdown
 */
export async function renderMarkdown(markdown: string): Promise<string> {
  return await marked.parse(markdown, {
    async: true,
    gfm: true,
  });
}
const FRONTMATTER_REGEX = /^---[\s\S]*?---/gm;

/**
 * Attempt to parse YAML frontmatter from a mixed yaml/md doc
 * @param content The raw markdown content
 * @returns Any successfully parsed frontmatter
 */
export function extractFrontmatter(content: string) {
  // If it does not start with `---`, it is invalid for frontmatter
  if (content.trim().indexOf("---") !== 0) {
    return {};
  }

  if (FRONTMATTER_REGEX.test(content)) {
    const frontmatterString = content.match(FRONTMATTER_REGEX)?.[0] ?? "";
    const cleanFrontmatter = frontmatterString.replaceAll("---", "").trim();
    return YAML.parse(cleanFrontmatter);
  }

  return {};
}

/**
 * Given a string of a markdown document, extract the markdown content
 * @param content The raw markdown content
 * @returns The markdown content without frontmatter
 */
export function extractMarkdown(content: string): string {
  if (content.trim().indexOf("---") !== 0) {
    return content;
  }
  return content.replace(FRONTMATTER_REGEX, "").trim();
}
